/* loader.h */
/*
Copyright 2002 John Coffman.
All rights reserved.

Licensed under the terms contained in the file 'COPYING' in the 
source directory.

*/
#include "common.h"

extern BUILTIN_FILE First;
extern BUILTIN_FILE Second;
extern BUILTIN_FILE Third;
extern BUILTIN_FILE Bitmap;
extern BUILTIN_FILE Mbr;
extern BUILTIN_FILE Chain;
extern BUILTIN_FILE Os2_d;

/* end loader.h */
