#if 0
; graph.S  is
Copyright 2000 John Coffman.
All rights reserved.

Licensed under the terms contained in the file 'COPYING' in the 
source directory.

#endif

#ifndef GRAPH_S
#define GRAPH_S


NS	= 179	; 0xB3
NSW	= 180
NSWW	= 181
NNSSW	= 182
SSW	= 183
SWW	= 184	; 0xB8
NNSSWW	= 185
NNSS	= 186
SSWW	= 187
NNWW	= 188
NNW	= 189
NWW	= 190
SW	= 191
NE	= 192	; 0xC0
NEW	= 193
SEW	= 194
NSE	= 195
EW	= 196
NSEW	= 197
NSEE	= 198
NNSSE	= 199
NNEE	= 200	; 0xC8
SSEE	= 201
NNEEWW	= 202
SSEEWW	= 203
NNSSEE	= 204
EEWW	= 205
NNSSEEWW = 206
NEEWW	= 207
NNEW	= 208	; 0xD0
SEEWW	= 209
SSEW	= 210
NNE	= 211
NEE	= 212
SEE	= 213
SSE	= 214
NNSSEW	= 215
NSEEWW	= 216
NW	= 217
SE	= 218	; 0xDA

#endif
; end of graph.S

